import logging
from asyncio.exceptions import CancelledError
from datetime import datetime

import pandas as pd
from IPython.display import clear_output
from dask.distributed import Client
from tabulate import tabulate

import dask
import time
from stratosphere import config
from stratosphere.run import Run
from stratosphere.utils import log
from typing import Dict, Union, Callable


def log_event(msg: Union[Dict, str]):
    """Log structured event with Dask.

    Args:
        msg (Union[Dict, str]): Event to log.
    """
    dask.distributed.get_worker().log_event("stratosphere", msg)


def dask_logger(run: Run, insert_delay: float = 0) -> Run:
    """Dask logger function that can be appended on experiments,
    to log the attributes of executed runs with Dask, enabling
    real-time experiment tracking. Tracked properties:

    Args:
        run (Run): Run to track.
        insert_delay (float, optional): Introduce a delay (seconds),
        useful to demonstrate the tracking functionality. Defaults to 0.

    Returns:
        Run: unaltered run after tracking it.
    """
    if insert_delay > 0:
        time.sleep(insert_delay)
    log_event({"run_id": run.run_id, "attributes": run.attributes})
    return run


@log.default_exception_handler
def monitor_latest_event(client_address: str = config.dask.scheduler_address):
    """Utility function to report the latest tracked Run event, useful to start
    debugging execution issues. The function returns after a keyboard interrupt
    or after the connection with the scheduler is lost, which can happen if:
    1. The connection delay is too low (see config.dask)
    2. The Dask scheduler terminated, either nicely or after a failure

    Args:
        client_address (str, optional): Dask scheduler address. Defaults to config.dask.scheduler_address.
    """
    log.init_logging(logging.INFO)

    def display_progress(df):
        clear_output(wait=True)
        print(
            f"Time: {datetime.now().strftime('%d/%m/%Y %H:%M:%S')} Count events:"
            f" {len(df)}\n"
        )
        print(tabulate(df.iloc[-1].to_frame()))

    monitor_events(display_progress, client_address=client_address)


@log.default_exception_handler
def monitor_events(func: Callable, client_address: str = config.dask.scheduler_address):
    """Utility function to execute a callable, passing a Pandas dataframe as argument, representing
    all the tracked events. It can be used to implement custom tracking monitors.

    Args:
        func (Callable): Function that accepts as argument a Pandas dataframe, representing the
        tracked events.
        client_address (str, optional): _description_. Defaults to config.dask.scheduler_address.
    """
    try:
        with Client(
            address=client_address, timeout=config.dask.client_timeout
        ) as client:
            while True:
                events = client.get_events("stratosphere")
                df = pd.json_normalize([event[1] for event in events])
                func(df)
                time.sleep(1)
    except (CancelledError, OSError):
        logging.error("Connection failed or cluster terminated")



import logging
import multiprocessing

import joblib

try:
    # We expect this import to work out if we want to use Dask.
    # Otherwise, not necessary.
    from distributed import get_client
except ImportError:
    pass

from stratosphere.utils.progress import progress

from stratosphere import config
from typing import Callable
from typing import List

from stratosphere.run import Run


class ProgressParallel(joblib.Parallel):
    """This class managed a progress bar monitoring the parallel execution of tasks."""

    def __init__(self, use_tqdm: bool = True, total: int = None, *args, **kwargs):
        """Create a new progress bar.

        Args:
            use_tqdm (bool, optional): If True, will use tqdm. IF False,
            the tasks will be executed with no progress bar. Defaults to True.
            total (int, optional): Total number of tasks planned for execution.
            Defaults to None.
            args: Additional args to pass to the joblib.Parallel constructor.
            kwargs: Additional kwargs pass to the joblib.Parallel constructor.
        """
        self._use_tqdm = use_tqdm
        self._total = total
        super().__init__(*args, **kwargs)

    def __call__(self, *args, **kwargs):
        with progress(disable=not self._use_tqdm, total=self._total) as self._pbar:
            return joblib.Parallel.__call__(self, *args, **kwargs)

    def print_progress(self) -> None:
        if self._total is None:
            self._pbar.total = self.n_dispatched_tasks
        self._pbar.n = self.n_completed_tasks
        self._pbar.refresh()


def parallel(funcs: List[Callable], n_jobs: int = -1) -> List[object]:
    """Execute a list of functions in parallel, returning their return values as a list.

    Args:
        funcs (List[Callable]): Functions to execute, no parameters. You can
        use functools.partial to build suitable functions.
        n_jobs (int, optional): Level of parallelization, passed to Joblib. Defaults to -1.

    Returns:
        List[object]: List of return values of the executed functions.
    """

    # We leave -1 as default value for n_jobs, so that Joblib will figure out a reasonable value.
    logging.info(f"Executing {len(funcs)} tasks on {n_jobs} parallel workers")

    # Instantiate the parallel executor
    p = ProgressParallel(n_jobs=n_jobs, use_tqdm=config.tqdm.enable, total=len(funcs))

    # Get the reeturn values, and return them as a lsit.
    # This function completes once all tasks return.
    rets = p(joblib.delayed(func)() for func in funcs)
    return rets


class Job:
    """This class is the entry point for all things execution."""

    def __init__(
        self,
        tasks: List[Callable] = None,
        n_jobs: int = -1,
        backend: str = joblib.parallel.DEFAULT_BACKEND,
    ):
        """Prepare a new job to execute.

        Args:
            tasks (List[Callable], optional): List of functions to be executed. Defaults to None.
            n_jobs (int, optional): Number of tasks to execute in parallel, same meaning as in
            Joblib. Defaults to -1.
            backend (str, optional): Backend to use, same meaning as in Joblib.
            Defaults to joblib.parallel.DEFAULT_BACKEND.
        """

        self.tasks = tasks
        self.n_jobs = n_jobs
        self.backend = backend

    def execute(self) -> List[Run]:
        """Execute the plan.

        Returns:
            List[Run]: List of runs, after being passed as
            input to the functions.
        """

        logging.info(f"Using backend: {self.backend}")

        n_jobs = self.n_jobs

        if self.backend == "dask":
            client = get_client()
            logging.info(f"Dashboard: {client.dashboard_link}")
            logging.info(f"Scheduler: {client.scheduler.address}")
            if self.n_jobs == -1:
                n_jobs = len(client.scheduler_info()["workers"])
        elif self.backend == joblib.parallel.DEFAULT_BACKEND:
            if self.n_jobs == -1:
                n_jobs = multiprocessing.cpu_count()

        with joblib.parallel_backend(self.backend):
            runs = parallel(self.tasks, n_jobs=n_jobs)

            # make sure that we have all the runs expect.
            assert len(runs) == len(self.tasks)

            return runs



import warnings

# Detect if we're running the notebook in a colab.
# If this is the case, we can avoid using tqdm.auto, which doesn't seem to work properly.
try:
    import google.colab

    IN_COLAB = True
except ModuleNotFoundError:
    IN_COLAB = False


with warnings.catch_warnings():
    warnings.simplefilter("ignore")

    if IN_COLAB:
        from tqdm import tqdm
    else:
        from tqdm.auto import tqdm

import sys


def progress(*args, **kwargs):
    if "pyodide" in sys.modules:
        # running in Pyodide
        kwargs["monitor_interval"] = 0
        tqdm.monitor_interval = 0

    return tqdm(*args, **kwargs)



import copy
from typing import Callable, List, Union


from stratosphere import config
from stratosphere.experiment import Experiment
from stratosphere.utils.enums import IfExists, enforce_enum
from stratosphere.store.db.database import Database
from stratosphere.utils import log
import pandas as pd


class Stratosphere:
    """Instantiate the Stratosphere handler, it can then be sued
    to create, query, manage experiments. You can instantiate one or more,
    they do work together nicely without interfering.
    """

    def __init__(self, url: str = config.db.default_url, echo: bool = False):
        """Create a new Stratosphere handler.

        Args:
            url (str, optional): Database URL. Defaults to config.db.default_url.
            echo (bool, optional): Enable SQLAlchemy logging. Defaults to False.
        """

        log.init_logging(config.log.level)
        self.db = Database(url, echo=echo)

    @log.default_exception_handler
    def define(
        self,
        name: str = None,
        funcs: Union[Callable, List[Callable]] = None,
        kwargs: dict = None,
        attributes: dict = None,
        parameter_grid: List = None,
    ) -> Experiment:
        """Define a new experiment.

        Args:
            name (str, optional): Name of the experiment (must be unique). Defaults to None.
            funcs (Union[Callable, List[Callable]], optional): Function(s) we want to apply to the input parameters. Defaults to None.
            kwargs (dict, optional): Fixed arguments to the functions. Defaults to None.
            attributes (dict, optional): Fixed experiment properties to be tracked. Defaults to None.
            parameter_grid (List, optional): Variable parameters to pass to the functions. Defaults to None.

        Returns:
            Experiment: The newly defined experiment, reaady to be executed.
        """

        return Experiment(
            db=self.db,
            name=name,
            funcs=funcs,
            kwargs=kwargs,
            attributes=attributes,
            parameter_grid=parameter_grid,
        )

    @log.default_exception_handler
    def ls(self) -> pd.DataFrame:
        """List the tracked experiments.

        Returns:
            pd.DataFrame: Pandas dataframe.
        """
        return Experiment.ls(self.db)

    @log.default_exception_handler
    def load(self, name: str = None) -> Experiment:
        """Load an experiment from the database and return it.

        Args:
            name (str, optional): The name of the experiment. Defaults to None.

        Returns:
            Experiment: The loaded experiment.
        """
        return Experiment.load(self.db, name)

    @log.default_exception_handler
    def persist(
        self, experiment: Experiment, if_exists: IfExists = "fail"
    ) -> Experiment:
        """Persist the experiment to the database (not necessarily
        the one used to track it initially).

        Args:
            experiment (Experiment): Experiment object to be persisted.
            if_exists (IfExists, optional): Either "replace" of "fail" in case
            the experiment name is already present. Defaults to "fail".

        Returns:
            Experiment: The persisted experiment.
        """

        # Enforce if_exists value
        if_exists = enforce_enum(if_exists, IfExists)

        # We make a copy of the experiment, and change the database reference.
        # We can then use all its methods, including persist(...).
        experiment = copy.deepcopy(experiment)
        experiment.db = self.db
        experiment.persist(if_exists=if_exists)
        return experiment



import logging
import random
from typing import Callable, List, Union

import joblib
import pandas as pd
from sqlalchemy.orm import load_only
from stratosphere.job import Job
from stratosphere.run import Run
from stratosphere.store import serialization
from stratosphere.store.db import schema
from stratosphere.store.db.database import Database, next_ulid, pandas_query
from stratosphere.utils import log, time
from stratosphere.utils.enums import IfExists, enforce_enum


class ExperimentAlreadyExists(Exception):
    """Exception to handle the attempt to add one more experiment with the same name."""

    def __init__(self, name):
        self.message = (
            f"Experiment '{name}' already existing, use if_exists=\"replace\" to"
            " overwrite it."
        )
        super().__init__(self.message)


class ExperimentNotFoundException(Exception):
    """Exception raised if the experiment is not found."""

    def __init__(self, name):
        self.message = (
            f"Experiment '{name}' not found. List with .ls() the available experiments."
        )
        super().__init__(self.message)


class RunNotFoundException(Exception):
    """Exception raised if the run is not found."""

    def __init__(self, run_id):
        self.message = f"Run '{run_id}' not found. List with .df() the available runs."
        super().__init__(self.message)


class Experiment:
    """This class represents a new experiment.

    Raises:
        ExperimentNotFoundException: Raised if we try to use a non-existing experiment.
        ExperimentAlreadyExists: Raised if the experiment name is already present.

    Returns:
        Experiment: A defined experiment, ready to be executed, tracked, stored, queried.
    """

    # schema_cls is the SQLAlchemy model mapped to this class.
    schema_cls = schema.Experiment

    def __init__(
        self,
        db: Database = None,
        name: str = None,
        funcs: Union[Callable, List[Callable]] = None,
        kwargs: dict = None,
        attributes: dict = None,
        parameter_grid: List = None,
        experiment_id: str = None,
        execution_time: int = None,
        completion_time: int = None,
        runs: List[Run] = None,
    ):
        """Create a new experiment.

        Args:
            db (Database, optional): Database to link to. Defaults to None.
            name (str, optional): Name of the experiment. Defaults to None.
            funcs (Union[Callable, List[Callable]], optional): One or more functions to call. Defaults to None.
            kwargs (dict, optional): Fixed parameters passed to the functions. Defaults to None.
            attributes (dict, optional): Fixed experiment attributes to be tracked. Defaults to None.
            parameter_grid (List, optional): Grid of parameters, used to generate the run parameters.
            Defaults to None.
            experiment_id (str, optional): ID of the experiment to be used. Defaults to None.
            execution_time (int, optional): Start time of the execution. Defaults to None.
            completion_time (int, optional): Completion time of the execution. Defaults to None.
            runs (List[Run], optional): List of runs belonging to the experiment. Defaults to None.
        """
        self.db = db
        self.name = name
        # normalise funcs to a list of functions.
        self.funcs = [funcs] if type(funcs) == callable else funcs
        self.kwargs = kwargs

        # we set the dict here and not directly in the constructor definition to make sure that
        # we create a new dict at every object, without sharing it as a static attribute across
        # instances.
        self.attributes = {} if attributes is None else attributes
        self.parameter_grid = parameter_grid

        # Setted here, for similar reasons of self.attributes.
        self.experiment_id = experiment_id if experiment_id is not None else next_ulid()
        self.execution_time = execution_time
        self.completion_time = completion_time
        self.runs = runs

    def __reduce__(self):
        """
        When we pickle the experiment (this happens as soon as we try to
        store it in the database), pickle will use this constructor to build
        a clean object to be stored.

        The .db field is set to None and is populated by the Stratosphere instance.
        This lets us manage more transparently the linked database, avoiding
        unpickable objects in SQLAlchemy.

        Returns:
            Experiment: An unpickled Experiment object, ready to be linked
            to a database and used.
        """
        return self.__class__, (
            None,
            self.name,
            self.funcs,
            self.kwargs,
            self.attributes,
            self.parameter_grid,
            self.experiment_id,
            self.execution_time,
            self.completion_time,
            self.runs,
        )

    @classmethod
    def load(cls, db: Database, name: str):
        """Load an experiment from database.

        Args:
            db (Database): Reference to a database.
            name (str): Name of the experiment.

        Raises:
            ExperimentNotFoundException: Raised if the experiment is not found.

        Returns:
            Experiment: Loaded experiment.
        """

        # Create a new session
        with db.session() as session:
            record = session.query(cls.schema_cls).filter_by(name=name).first()

            # IF the record is not found, raise an error.
            if record is None:
                raise ExperimentNotFoundException(name)

            # Unpicke the serialized object.
            experiment = serialization.pickle_loads(record.serialized)
            # Set the db of the experiment. Pickled db instances are never re-instantiated
            # completely, this let us reuse the existing database instance. See the Database
            # class for more comments on this.
            #
            # Important: this let us use SQLite memory databases, as only one instance is
            # used on all experiments.
            experiment.db = db
            return experiment

    @log.default_exception_handler
    @log.timeit
    def execute(
        self,
        funcs: Union[str, Callable, List[Callable]] = None,
        backend=joblib.parallel.DEFAULT_BACKEND,
        n_jobs=-1,
    ):
        """Execute the experiment, then retuning it.

        Args:
            funcs (Union[Callable, List[Callable]], optional): Function(s) that will override the defined ones.
            If the value is "all", the re-evaluation of all defined functions is forced. Defaults to None.
            backend (_type_, optional): Joblib baackend. Defaults to joblib.parallel.DEFAULT_BACKEND.
            n_jobs (int, optional): Parallelization degree, defined as in Joblib. Defaults to -1.

        Returns:
            Run: the same experiment, ready for chained operations.
        """

        self.execution_time = time.current_time()

        if funcs == "all":
            # If we pass 'all', we force the re execution of all funcs.
            funcs = self.funcs

        # If this is not the first execution ...
        if self.runs is not None:
            # If we want to override the functions to execute ...
            if funcs is not None:
                # ... we can reuse the runs context, and execute the functions on them.
                tasks = [
                    Run.execute_func(run, override_funcs=funcs) for run in self.runs
                ]

            else:
                # else: ... there is nothing to do, the runs context is already up to date.
                tasks = []
        else:
            # If this is the first execution, we build and execute the runs from scratch.
            parameters_list = self.parameter_grid.copy()
            # We shuffle the parameters, so that partial experiment statistics and expected
            # duration are more representative of all runs.
            random.shuffle(parameters_list)

            tasks = [
                Run.execute_func(
                    Run(
                        kwargs=self.kwargs,
                        parameters=parameters,
                        funcs=self.funcs,
                        experiment_name=self.name,
                    ),
                    override_funcs=funcs,
                )
                for parameters in parameters_list
            ]

        if len(tasks) > 0:
            # We update the runs context only if there are tasks to execute.
            self.runs = Job(tasks, n_jobs=n_jobs, backend=backend).execute()

        self.completion_time = time.current_time()
        return self

    def record(self) -> schema.Experiment:
        """Build an SQLAlchemy ORM object from the existing Experiment object.

        Returns:
            schema.Experiment: SQLAlchemy ORM object.
        """
        return self.schema_cls(
            id=self.experiment_id,
            name=self.name,
            execution_time=self.execution_time,
            completion_time=self.completion_time,
            attributes=serialization.json_dumps({**self.attributes}),
            serialized=serialization.pickle_dumps(self),
        )

    def info(self):
        """Print some stats about the experiment.

        Returns:
            Experiment: self
        """
        logging.info("Experiment info")
        logging.info(f"  id...............: {self.experiment_id}")
        logging.info(f"  name.............: '{self.name}'")
        logging.info(f"  funcs............: {[func.__name__ for func in self.funcs]}")
        logging.info(f"  kwargs...........: {list(self.kwargs.keys())}")

        logging.info(
            "  parameter_grid...:"
            f" {list(self.parameter_grid[0].keys())} ({len(self.parameter_grid)})"
        )

        logging.info(f"  attributes.......: {list(self.attributes.keys())}")
        logging.info(f"  size (Mb)........: {self.size('mb')} (compressed)")

        return self

    @log.default_exception_handler
    def persist(self, if_exists: IfExists = IfExists["fail"]):
        """Persist the experiment.

        Args:
            if_exists (IfExists, optional): Either "replace" or "fail". Defaults to "fail".

        Returns:
            _type_: self.
        """

        logging.info("Persisting experiment")
        if_exists = enforce_enum(if_exists, IfExists)

        # First, we try to delete the record, honoring the if_exists preference.
        self.delete(if_exists)
        # Then, we:
        # 1. Add the experiment record to the "experiments" table.
        # 2. Add the experiment table "experiment_{name}".

        with self.db.session() as session:
            session.add(self.record())
            session.commit()
        self.db.pandas_to_sql(self.df(), f"experiment_{self.name}", if_exists.name)

        return self

    @log.default_exception_handler
    def lookup(self, run_id: str) -> Run:
        """Lookup a run by ID

        Args:
            run_id (str): ID of the run to lookup.

        Raises:
            RunNotFoundException: Raised if run ID not found.

        Returns:
            Run: The matching run.
        """

        if self.runs is None:
            raise RunNotFoundException(run_id)

        # A rather unperformance way to look up the run.
        # TODO: change .runs to a dict.
        for run in self.runs:
            if run.run_id == run_id:
                return run

    @log.default_exception_handler
    def delete(self, if_exists: IfExists = IfExists["fail"]) -> None:
        """Delete the expeirment.

        Args:
            if_exists (IfExists, optional): Either "replace" or "fail". In case
            the experiment exists and the value is "replace", nothing happens.
            If the experiment exists and the value is "fail", it will raise an
            exception. It is designed to work correctly together with insertions.
            Defaults to "fail".

        Raises:
            ExperimentAlreadyExists: Raised if the experiment exists.
        """

        if_exists = enforce_enum(if_exists, IfExists)

        # Create a new session
        with self.db.session() as session:
            # If we find the record ...
            if session.query(self.schema_cls).filter_by(name=self.name).count() > 0:
                # And we are fine deleting it, proceed.
                if if_exists == IfExists["replace"]:
                    session.query(Experiment.schema_cls).filter(
                        Experiment.schema_cls.name == self.name
                    ).delete()
                else:
                    # Otherwise, raise an exception.
                    raise ExperimentAlreadyExists(self.name)
            session.commit()

        # We also need to drop the experiment table.
        self.db.drop_table(f"experiment_{self.name}")

    @log.default_exception_handler
    def df(self, include_experiment=True) -> pd.DataFrame:
        """Returns a Pandas dataframe representing the experiment, including
        parameters and attributes. The processing does not depend on
        database queries, but the returned dataframe matches the contents
        of the experiment table (which is generated using this method).

        In presence of overlapping column names from the fixed attributes
        of the experiment and the run attributes, the run_attributes use
        "_run" as suffix.

        Args:
            include_experiment (bool, optional): If False, do not include the
            fixed experiment attributes (it will be a constant column).
            Defaults to True.

        Returns:
            pd.DataFrame: Pandas dataframe representing the tracked properties.
        """

        # Load the dataframe from a dict or list of dicts.
        df_runs = pd.json_normalize(
            [{**run.attributes, **{"id_run": run.run_id}} for run in self.runs]
        )

        if not include_experiment:
            return reorder_columns(df_runs, ["id_run"]).apply(
                pd.to_numeric, errors="ignore"
            )
        else:
            df_experiment = pd.json_normalize(
                {
                    **self.attributes,
                    **{"id_experiment": self.experiment_id, "name": self.name},
                }
            )

            # In case of overlapping column names, the ones from runs use the _run suffix.
            df = df_experiment.merge(
                df_runs,
                how="cross",
                suffixes=("", "_run"),
            )

            return reorder_columns(df, ["name", "id_experiment", "id_run"]).apply(
                pd.to_numeric, errors="ignore"
            )

    def size(self, unit: str = "b") -> int:
        """Return the size of the pickled version of the expeirment.

        Args:
            unit (str, optional): Unit of measure: "b" (Bytes), "kb" (KiloBytes), "mb" (MegaBytes). Defaults to "b".

        Returns:
            int: Size of the experiment once pickled.
        """
        return serialization.pickle_size(self, unit=unit)

    @classmethod
    def ls(cls, db: Database) -> pd.DataFrame:
        """List experiments in a database.

        Args:
            db (Database): Database instance.

        Returns:
            pd.DataFrame: List of experiments.
        """
        with db.session() as session:
            df_experiments = pandas_query(
                session.query(Experiment.schema_cls).options(
                    load_only(
                        Experiment.schema_cls.id,
                        Experiment.schema_cls.name,
                        Experiment.schema_cls.attributes,
                    )
                ),
                session,
            )

            df_experiments = explode_json_column(df_experiments, "attributes").rename(
                columns={"id": "experiment_id"}
            )

            return df_experiments


def explode_json_column(
    df: pd.DataFrame, col_name: str, suffix: str = None
) -> pd.DataFrame:
    """Explode a column in the Pandas dataframe containing a dict to a list of columns.
    Useful to handle the "attributes" column in the "Experiments" table, which contains
    the serialized values as JSON.

    Args:
        df (pd.DataFrame): Pandas dataframe
        col_name (str): Column to process.
        suffix (str, optional): In case exploded columns already exist, use this suffix.
        Defaults to _{col_name}.

    Returns:
        pd.DataFrame: Resulting Pandas dataframe.
    """

    if suffix is None:
        suffix = f"_{col_name}"

    # Explode column containing json to multiple columns, in a dataframe.
    df_exploded = pd.json_normalize(df[col_name].apply(serialization.json_loads))

    return df.drop(columns=[col_name]).merge(
        df_exploded,
        how="left",
        left_index=True,
        right_index=True,
        suffixes=("", suffix),
    )


def reorder_columns(df: pd.DataFrame, ordered_columns: List[str]) -> List[str]:
    """Given a dataframe, enforce a partial ordering of some columns, appending
    the remaining ones in sorted order.

    Args:
        df (pd.DataFrame): Consider the columns from thsi dataframe.
        ordered_columns (List[str]): Consider this partial ordering of the columns.

    Returns:
        List[str]: New ordering of all columns for the {df} dataframe.
    """

    remaining_columns = sorted(
        [col_name for col_name in df.columns if col_name not in ordered_columns]
    )
    return df[ordered_columns + remaining_columns]
